import docker
from os import path


def build_image():
    client = docker.from_env()
    current_dir = path.dirname(__file__)
    client.images.build(
        tag='image-fixture', path=path.join(current_dir, 'fixtures'))



from slipway.container import Container
from test.util import build_image
import docker
import docker.errors
from subprocess import Popen, TimeoutExpired
import pty
import os


class FakeArgs(object):
    def __init__(self, workspace):
        self.image = 'image-fixture'
        self.volume = []
        self.environment = []
        self.workspace = workspace


def create_container(tmp_path):
    build_image()
    workspace = tmp_path / 'workspace'
    args = FakeArgs(str(workspace))
    client = docker.from_env()
    try:
        container = client.containers.get('slipway_image_fixture')
        container.kill()
    except docker.errors.NotFound:
        pass
    master_fd, slave_fd = pty.openpty()
    process = Popen(
        Container(client, args)._run_arguments(),
        preexec_fn=os.setsid,
        stdout=slave_fd,
        stderr=slave_fd,
        stdin=slave_fd
        )
    try:
        process.wait(timeout=2)
        assert process.returncode is None
    except TimeoutExpired:
        pass
    return client.containers.get('slipway_image_fixture')


def test_run(tmp_path):
    container = create_container(tmp_path)
    code, output = container.exec_run('ps aux')
    assert code == 0
    assert 'tmux' in str(output)
    container.kill()


def test_entrypoint(tmp_path):
    container = create_container(tmp_path)
    code, output = container.exec_run('stat -c %U /test-volume')
    assert code == 0
    assert 'foobar' in str(output)
    container.kill()



from test.util import build_image
from slipway.image import Image
from docker.errors import ImageNotFound
import docker

client = docker.from_env()


class FakeArgs(object):
    def __init__(self, image):
        self.image = image


def test_initialize():
    try:
        client.images.get('busybox')
        client.images.remove('busybox')
    except ImageNotFound:
        pass
    args = FakeArgs('busybox')
    image = Image(client, args)
    image.initialize()
    assert client.images.get('busybox') is not None


def test_volumes():
    build_image()
    image = Image(client, FakeArgs('image-fixture'))
    assert len(image.volumes) == 1
    assert image.volumes[0] == '/test-volume'


def test_entrypoint():
    build_image()
    image = Image(client, FakeArgs('image-fixture'))
    assert image.entrypoint == ['/entrypoint.sh']


def test_user():
    build_image()
    image = Image(client, FakeArgs('image-fixture'))
    assert image.user == 'foobar'


def test_home():
    build_image()
    image = Image(client, FakeArgs('image-fixture'))
    assert image.home == '/home/foobar'



import re


def snake_case(text):
    """
    Used for converting paths and such to a snake case format.
    """
    no_symbol = re.sub('[^A-Za-z0-9]', '_', text)
    no_duplicate = re.sub('[_]{2,}', '_', no_symbol.lower())
    return re.sub('(^_|_$)', '', no_duplicate)



from docker.errors import ImageNotFound

from multiprocessing.dummy import Pool
import math
from os import path
import re
import requests
from sys import stdout


# ANSI (terminal) escape sequences
CLEAR_LINE = '\x1b[0K'
SHOW_CURSOR = '\x1b[?25h'
HIDE_CURSOR = '\x1b[?25l'


class Image(object):
    def __init__(self, client, args):
        self.client = client
        self.args = args
        self.name = args.image
        self._image = None

        colon_parts = self.name.split(':')
        if len(colon_parts) == 1:
            self._tag = 'latest'
        else:
            self._tag = colon_parts[1]

        slash_parts = colon_parts[0].split('/')

        if len(slash_parts) == 3:
            self._registry_base_url = slash_parts[0]
        else:
            self._registry_base_url = 'https://index.docker.io'

        if len(slash_parts) == 1:
            self._repository = 'library/' + slash_parts[0]
        elif len(slash_parts) == 2:
            self._repository = '/'.join(slash_parts)
        else:
            self._repository = '/'.join(slash_parts[:1])

    def _docker_image(self):
        if self._image is None:
            self._image = self.client.images.get(self.name)
        return self._image

    def stale(self):
        """
        Returns true if the image isn't in sync with what is in the registry.
        """
        distribution = self.client.api.inspect_distribution(self.name)
        remote_digest = distribution['Descriptor']['digest']
        for local_digest in self._docker_image().attrs['RepoDigests']:
            remote_part = remote_digest.split(':')[1]
            local_part = local_digest.split(':')[1]
            if remote_part != local_part:
                return True
        return False

    def _registry_auth_url(self, base_url):
        if base_url == 'https://index.docker.io':
            return ('https://auth.docker.io/token', 'registry.docker.io')
        else:
            response = requests.get(
                base_url + '/v2/_catalog')
            if response.status_code != 401:
                message = 'Unexpected status code while determining ' + \
                    'authentication url'
                raise Exception(message)
            authenticate = response.headers['Www-authenticate']
            auth_url = re.match('realm="([^,]+)"', authenticate)
            service_url = re.match('service="([^,]+)"', authenticate)
            return (auth_url, service_url)

    def _urljoin(self, *parts):
        return '/'.join(parts)

    def _registry_token(self, auth_url, service_url):
        scope = 'repository:' + self._repository + ':pull'
        query = 'scope=' + scope + '&service=' + service_url
        url = auth_url + '?' + query
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(
                'Failed to retrieve auth token')
        return response.json()['token']

    def _registry_manifest(self, token):
        url = self._urljoin(
            self._registry_base_url, 'v2', self._repository, 'manifests',
            self._tag)
        headers = {'authorization': 'Bearer ' + token}
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            message = 'Unexpected {} response retrieving manifest'
            formatted = message.format(response.status_code)
            raise Exception(formatted)
        return response.json()

    def _manifest_layer_size(self, token, manifest):
        def fetch(layer):
            digest = layer['blobSum']
            url = self._registry_base_url + '/v2/' + self._repository + \
                '/blobs/' + digest
            headers = {'authorization': 'Bearer ' + token}
            with requests.get(url, headers=headers, stream=True) as response:
                if response.status_code != 200:
                    message = 'Unexpected {} response determining layer size'
                    raise Exception(message.format(response.status_code))
                content_length = int(response.headers['content-length'])
                id = digest.replace('sha256:', '')
                return (id, content_length)
        pool = Pool(10)
        sizes = pool.map(fetch, manifest['fsLayers'])
        layer_sizes = {}
        total = 0
        for (id, size) in sizes:
            total += size
            layer_sizes[id] = size
        return (total, layer_sizes)

    def _format_bytes(self, unformatted):
        mb = (unformatted / 1024) / 1024
        rounded = math.floor(mb * 10) / 10
        return '{} MB'.format(rounded)

    def _print_pull_status(self, total, downloaded, extracted):
        formatted_total = self._format_bytes(total)
        if self._pull_status_started:
            stdout.write(CLEAR_LINE)
            stdout.flush()
        else:
            self._pull_status_started = True

        stdout.write('Downloaded: {} / {}, Extracted: {} / {}\r'.format(
            self._format_bytes(downloaded),
            formatted_total,
            self._format_bytes(extracted),
            formatted_total))
        stdout.flush()

    def _consume_pull_status_stream(self, total, layer_sizes, stream):
        downloaded = 0
        extracted = 0
        downloaded_layers = {}
        extracted_layers = {}
        for status in stream:
            if status['status'] == 'Downloading':
                detail = status['progressDetail']
                if status['id'] in downloaded_layers:
                    layer = downloaded_layers[status['id']]
                    downloaded -= layer['current']
                downloaded += detail['current']
                downloaded_layers[status['id']] = detail
                self._print_pull_status(total, downloaded, extracted)
            elif status['status'] == 'Extracting':
                if status['id'] in extracted_layers:
                    layer = extracted_layers[status['id']]
                    extracted -= layer['current']
                extracted += detail['current']
                extracted_layers[status['id']] = detail
                self._print_pull_status(total, downloaded, extracted)
            elif status['status'] == 'Already exists':
                size = None
                for layer in layer_sizes:
                    if layer.startswith(status['id']):
                        size = layer_sizes[layer]
                        break
                if size is None:
                    raise Exception('Unexpected layer from daemon')
                downloaded += size
                extracted += size
                self._print_pull_status(total, downloaded, extracted)
        pass

    def pull(self):
        print('Computing total image size')
        (auth_url, service_url) = self._registry_auth_url(
            self._registry_base_url)
        token = self._registry_token(auth_url, service_url)
        manifest = self._registry_manifest(token)
        (total, layer_sizes) = self._manifest_layer_size(token, manifest)
        self._pull_status_started = False
        print('Downloading layers')
        stream = self.client.api.pull(
            self._repository, tag=self._tag, stream=True, decode=True)
        try:
            stdout.write(HIDE_CURSOR)
            stdout.flush()
            self._consume_pull_status_stream(total, layer_sizes, stream)
        finally:
            stdout.write(SHOW_CURSOR)
            stdout.flush()

    def exists(self):
        """
        Returns true if the image exists locally, false otherwise.
        """
        try:
            self._docker_image()
            return True
        except ImageNotFound:
            return False

    def initialize(self):
        """
        Pulls the image if not present
        """
        if not self.exists():
            message = 'Image {} not found, attempting to pull down'
            print(message.format(self.name))
            self.pull()
        elif self.args.pull:
            print('Checking for updates')
            if self.stale():
                print('Updates available, pulling')
                self.pull()

    @property
    def volumes(self):
        """
        Returns all volumes tied to the image's container configuration
        """
        config = self._docker_image().attrs['ContainerConfig']
        if config['Volumes'] is None:
            return []
        return list(config['Volumes'].keys())

    @property
    def entrypoint(self):
        return self._docker_image().attrs['ContainerConfig']['Entrypoint']

    @property
    def user(self):
        return self._docker_image().attrs['ContainerConfig']['User']

    @property
    def home(self):
        return path.join('/home', self.user)



# from setuptools.command.install import install
from setuptools import setup, find_packages

setup(
    name='slipway',
    packages=find_packages(),
    version='0.6.0',
    description='CLI tool for managing development containers',
    author='Jonathan Boudreau',
    author_email='jonathan.boudreau.92@gmail.com',
    url='https://github.com/AGhost-7/slipway',
    download_url='',
    keywords=['docker', 'development'],
    classifiers=[],
    install_requires=[
        'docker >= 3.0.0'
    ],
    entry_points={
        'console_scripts': [
            'slipway=slipway:main'
        ]
    }
)



import docker
from slipway.volumes import Volumes
from slipway.image import Image
from test.util import build_image

client = docker.from_env()
image_name = build_image()


class FakeArgs(object):
    def __init__(self, image):
        self.image = image


def test_volumes_init():
    for volume in client.volumes.list():
        if volume.name == 'slipway_image_fixture_test_volume':
            volume.remove()
    args = FakeArgs('image-fixture')
    image = Image(client, args)
    volumes = Volumes(client, args, image)
    volumes.initialize()
    matches = [
        volume
        for volume in client.volumes.list()
        if volume.name == 'slipway_image_fixture_test_volume'
    ]
    assert len(matches) == 1
    matches[0].remove()


def test_volumes_purge():
    args = FakeArgs('image-fixture')
    image = Image(client, args)
    volumes = Volumes(client, args, image)
    client.volumes.create('slipway_image_fixture_test_volume')
    volumes.purge()
    for volume in client.volumes.list():
        assert volume.name != 'slipway_image_fixture_test_volume'




import docker
from slipway.binds import Binds
from slipway.image import Image

client = docker.from_env()


class FakeArgs(object):
    def __init__(self, volume, workspace):
        self.image = 'image-fixture'
        self.volume = volume
        self.workspace = workspace


def test_initialize(tmp_path):
    volume = tmp_path / 'initialize'
    assert not volume.exists()
    args = FakeArgs(
        [str(volume) + ':/initialize'],
        str(tmp_path / 'workspace'))
    image = Image(client, args)
    binds = Binds(client, args, image)
    binds.initialize()
    assert volume.exists()


def test_list(tmp_path):
    file_bind = tmp_path / 'file.txt'
    file_bind.touch()
    dir_bind = tmp_path / 'directory'
    dir_bind.mkdir()
    host_workspace = tmp_path / 'some-workspace'
    host_workspace.mkdir()
    args = FakeArgs(
        [str(file_bind) + ':/file', str(dir_bind) + ':/dir'],
        str(host_workspace))
    image = Image(client, args)
    binds = Binds(client, args, image)
    matches = [
        bind for bind in binds.list()
        if bind.container_path in ('/file', '/dir')
    ]
    assert len(matches) == 2

    for bind in matches:
        if bind.container_path == '/file':
            assert bind.type == 'f'
        else:
            assert bind.type == 'd'

    workspace = [
        bind for bind in binds.list()
        if bind.container_path == '/home/foobar/some-workspace'
    ]
    assert len(workspace) == 1
